from dataclasses import dataclass, field


@dataclass
class Tested:
    """
    The results of testing every URL
    """

    has_new: set[str] = field(default_factory=set)
    ignore: set[str] = field(default_factory=set)
    skip: set[str] = field(default_factory=set)
    failed: set[str] = field(default_factory=set)
    unknown: set[str] = field(default_factory=set)



from __future__ import annotations
from typing import TYPE_CHECKING
from pathlib import Path
from time import sleep
import traceback

from manga.utils import extract_url, lsf

from .test_url import test_url
from .dispatch import dispatch
from .state import State, URL
from .status import Skipped
from .results import results

if TYPE_CHECKING:
    from collections.abc import Callable


_skipped = Skipped()


def evaluate_urls(urls: list[URL], update_pbar: Callable[[], None], skip: set[str], delay: int) -> None:
    """
    :param urls: The list of URLs to test
    :param update_pbar: A function that updates the progress bar to be called when a URL is tested
    :param skip: A set of domains to skip
    :param delay: How many seconds to sleep before testing successive urls
    """
    try:
        if urls[0].domain in skip:
            for i in urls:
                i.status = _skipped
                update_pbar()
            return
        first: bool = False
        for url in urls:
            if not first:
                sleep(delay)
                first = False
            url.status = test_url(url.url)
            update_pbar()
    except KeyboardInterrupt:
        raise
    except:
        traceback.print_exc()
        raise


def test_sites(directory: Path, skip: set[str] | list[str], opener: str, no_prompt: bool, delay: int) -> bool:
    """
    Test each file in directory, print the results open them as needed
    """
    if isinstance(skip, list):
        return test_sites(directory, set(skip), opener, no_prompt, delay)
    print("Checking arguments...")
    directory = directory.resolve()
    assert delay >= 0, "Delay may not be negative"
    assert directory.exists(), f"{directory} does not exist"
    assert directory.is_dir(), f"{directory} is not a directory"
    # Test
    print("Scanning files...")
    state = State({extract_url(i) for i in lsf(directory)})
    print(f"Testing {len(state)} urls...")
    dispatch(state, evaluate_urls, skip, delay)
    # Results
    results(state, no_prompt, opener)
    return True



from __future__ import annotations
from typing import TYPE_CHECKING, Protocol
from concurrent.futures import thread

from rich.progress import Progress, BarColumn, TextColumn, TimeRemainingColumn, MofNCompleteColumn, TaskProgressColumn

if TYPE_CHECKING:
    from concurrent.futures import Future
    from collections.abc import Callable
    from typing import Any

if TYPE_CHECKING:
    from .state import URL, State


class _ThreadHandler(thread.ThreadPoolExecutor):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.futures: list[Future] = []

    def add(self, fn: Callable[..., Any], *args: Any, **kwargs: Any) -> None:
        self.futures.append(super().submit(fn, *args, **kwargs))

    def kill(self):
        self.shutdown(wait=False)
        for i in self.futures:
            i.cancel()


class FuncType(Protocol):
    def __call__(self, urls: list[URL], update_pbar: Callable[[], None], *args: Any, **kwargs: Any) -> None: ...


def dispatch(state: State, func: FuncType, *args: Any, **kwargs: Any) -> None:
    buckets = state.domains()
    with Progress(
        TextColumn("[progress.description]{task.description}"),
        TaskProgressColumn(),
        BarColumn(None),
        MofNCompleteColumn(),
        TimeRemainingColumn(compact=True, elapsed_when_finished=True),
        # transient=True, TODO
        expand=True,
    ) as pbar:
        with _ThreadHandler(max_workers=len(buckets)) as executor:  # No DOS-ing
            for domain, urls in buckets.items():
                task = pbar.add_task(f"{domain}:", total=len(urls))
                executor.add(func, urls, lambda bound=task: pbar.update(bound, advance=1), *args, **kwargs)



from __future__ import annotations
from typing import TYPE_CHECKING, Protocol
from concurrent.futures import thread
import signal

if TYPE_CHECKING:
    from concurrent.futures import Future
    from typing import Any


class _SigintHandler(Protocol):
    def __call__(self, executor: ThreadHandler, *args: Any):
        pass


class ThreadHandler(thread.ThreadPoolExecutor):
    def __init__(self, sigint_handler: _SigintHandler, *args: Any, **kwargs: Any):
        self._handler = lambda *x: sigint_handler(self, *x)
        super().__init__(*args, **kwargs)
        self.futures: list[Future] = []

    def add(self, fn, *args: Any, **kwargs: Any) -> None:
        self.futures.append(super().submit(fn, *args, **kwargs))

    def kill(self):
        self.shutdown(wait=False)
        for i in self.futures:
            i.cancel()

    def __enter__(self) -> ThreadHandler:
        self._orig: Any = signal.signal(signal.SIGINT, self._handler)
        rv = super().__enter__()
        assert rv is self
        return self

    def __exit__(self, *args):
        super().__exit__(*args)
        signal.signal(signal.SIGINT, self._orig)



from __future__ import annotations
from collections.abc import Iterable
from pathlib import Path


__all__ = ("lsf", "mv")


def lsf(d: Path) -> tuple[Path, ...]:
    """
    ls all files in d recursively, but ignore likely undesired files
    """
    each: Iterable[Path] = (i.resolve() for i in d.resolve().rglob("*"))
    ignore: set[str] = {".DS_Store"}
    return tuple(i for i in each if i.is_file() and i.name not in ignore)


def mv(a: Path, b: Path, *, dryrun: bool = False) -> None:
    """
    Rename a to b
    """
    a = a.resolve()
    b = b.resolve()
    print(f"{a} --> {b}")
    if not dryrun:
        a.rename(b)



from __future__ import annotations
from typing import TYPE_CHECKING
from pathlib import Path
import os

import requests
import tqdm

from manga.utils import extract_url, lsf
from manga import sites

from .thread_handler import ThreadHandler
from .results import handle_results
from .tested import Tested

if TYPE_CHECKING:
    from typing import Any


mk_open_remaining_first: bool = True


def evaluate(url: str, tested: Tested, pbar: tqdm.std.tqdm) -> None:
    """
    Determine if url has a new chapter or not
    Store the result in tested and update pbar
    """
    try:
        status: bool = sites.test(url, timeout_retries=3, base_delay=30)
        (tested.has_new if status else tested.ignore).add(url)
    except sites.UnknownDomain:
        tested.unknown.add(url)
    except requests.exceptions.RequestException:
        tested.failed.add(url)
    except Exception as e:
        print(f"Thread failed: {e}")
    finally:
        pbar.update()


def open_new(directory: Path, skip: set[str] | list[str]) -> bool:
    """
    Open each file in directory that has a new chapter ready
    """
    if isinstance(skip, list):
        return open_new(directory, set(skip))
    print("Checking arguments...")
    directory = directory.resolve()
    assert directory.exists(), f"{directory} does not exist"
    assert directory.is_dir(), f"{directory} is not a directory"
    # Determine which requests must be made
    print("Scanning files...")
    urls: set[str] = {extract_url(i) for i in lsf(directory)}
    results = Tested()

    # Sigint handler
    def handler(executor: ThreadHandler, *_: Any) -> None:
        global mk_open_remaining_first  # pylint: disable=global-statement
        if not mk_open_remaining_first:
            os._exit(1)  # pylint: disable=protected-access
        mk_open_remaining_first = False
        print("Terminating executor...")
        executor.kill()
        handle_results(urls, results)
        os._exit(0)  # pylint: disable=protected-access

    # Determine what to open
    print(f"Making at most {len(urls)} requests...")
    with tqdm.tqdm(total=len(urls), dynamic_ncols=True) as pbar:
        with ThreadHandler(max_workers=32, sigint_handler=handler) as executor:
            for i in urls:
                if sites.get_domain(i) in skip:
                    results.skip.add(i)
                    pbar.update()
                else:
                    executor.add(evaluate, i, results, pbar)
    # Open links
    handle_results(urls, results)
    return True



from .guess import guess, trash as guess_trash, max_chapter as guess_max_chapter
from .inc_chapter import inc_chapter
from .open_new import open_new
from .test_sites import test_sites
from .unnumbered_helper import unnumbered_helper
from .up_number import up_number
from .no_unicode import no_unicode
